type op = Add | Sub | Mult | Div | Exp | Equal | Neq | Less | Leq | Greater | Geq

type prim = Int | Float | Bool | String

type basic = Array | Signal

type expr =
    Literal of int
  | Id of string * expr
  | String_literal of string
  | Float_literal of float
  | Bool_literal of bool
  | Binop of expr * op * expr
  | Call of string * expr list
  | Basic_Op of basic * string * int * string *  expr
  | Noexpr

type variable = {
    typ		:	prim;
	name	: 	string;
    exp		:	expr;
}

type vdecl =
	Variable_Dec of variable

type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | Prim_Assign of variable
  | While of expr * stmt
  | Sum of string * string * int * int * expr
  | Basic_Dec of string * basic * int
  | Basic_Sum of string * string * int * int * expr
  | Print of expr

type formal =
{
	form_type	: prim;
	form_name	: string;
}

type func_decl = 
{
    is_stream	: bool;
    fname 		: string;
    ret_type 	: prim option;
    formals 	: formal list;
    body 		: stmt list;
}

type program = vdecl list * func_decl list

(* Low-level AST printing, to help debug the structure.  These functions are
   only for debugging (the -r flag) and can be removed. *)

let rec expr_s = function
   Literal(l) -> "Literal " ^ string_of_int l
 | Id(s, e) -> "Id " ^ s ^ expr_s e
 | Binop(e1, o, e2) -> "Binop (" ^ expr_s e1 ^ ") " ^
       (match o with Add -> "Add" | Sub -> "Sub" | Mult -> "Mult" |
                     Div -> "Div" | Exp -> "Exp" | Equal -> "Equal" | Neq -> "Neq" |
                     Less -> "Less" | Leq -> "Leq" | Greater -> "Greater" |
                     Geq -> "Geq") ^ " (" ^ expr_s e2 ^ ")"
 | String_literal(s) -> "String Lit: " ^ s ^ "."
 | Float_literal(f) -> "Float lit: " ^ string_of_float f ^ "."
 | Bool_literal(b) -> "Bool lit: " ^ string_of_bool b ^ "."
 | Call(f, es) -> "Call " ^ f ^ " [" ^
        String.concat ", " (List.map (fun e -> "(" ^ expr_s e ^ ")") es) ^ "]"
 | Basic_Op(b, s, i, a, e) -> 
			(match b with Signal -> "Signal" | Array -> "Array") ^ " " ^
			s ^ (match b with Signal -> "[time - " | Array -> "[") 
			^ string_of_int i ^ "] = " ^ expr_s e
 | Noexpr -> ""

let prim_s = function
	Int -> "int"
  | Float -> "float"
  | String -> "string"
  | Bool -> "bool"

let string_of_prim = function
	Int -> "int"
  | Float -> "float"
  | String -> "std::string"
  | Bool -> "bool"

let string_of_type = function
	Some(x) -> string_of_prim x
  | _ -> ""

let string_of_basic = function
	Array -> "Array"
  | Signal -> "Signal"

let string_of_var var = 
	string_of_prim var.typ ^ " " ^ var.name ^ " " ^ expr_s var.exp

let string_of_vdecl = function
	Variable_Dec(v) -> string_of_var v

let rec stmt_s = function
   Block(ss) -> "Block [" ^ String.concat ",\n"
                             (List.map (fun s -> "(" ^ stmt_s s ^ ")") ss) ^ "]"
 | Expr(e) -> "Expr (" ^ expr_s e ^ ")"
 | Return(e) -> "Return (" ^ expr_s e ^ ")"
 | If(e, s1, s2) -> "If (" ^ expr_s e ^ ") (" ^ stmt_s s1 ^ ") (" ^
                                                stmt_s s2 ^ ")"
 | For(e1, e2, e3, s) -> "For (" ^ expr_s e1 ^ ") (" ^ expr_s e2 ^
                            ") (" ^ expr_s e3 ^ ") (" ^ stmt_s s ^ ")"
 | While(e, s) -> "While (" ^ expr_s e ^ ") (" ^ stmt_s s ^ ")"
 | Print(s) -> "Printing: " ^ expr_s s ^ "."
 | Basic_Dec(s, b, i) -> (match b with Signal -> "Signal" | Array -> "Array") ^ " " ^ s ^ " " ^ string_of_int i ^ "."
 | Prim_Assign(v) -> string_of_var v

let string_of_formal forml = 
	string_of_prim forml.form_type ^ " " ^ forml.form_name

let func_decl_s f =
  "{" 
  ^ (match f.is_stream with true -> "Stream" | false -> "") ^ 
  " return type = " ^ string_of_type f.ret_type ^ "fname = \"" ^ f.fname ^ "\"\n   formals = [" ^
  String.concat ", " (List.map string_of_formal f.formals) ^ "]\n" ^
  String.concat ",\n" (List.map stmt_s f.body) ^
  "]}\n"


let program_s (vars, funcs) = "([" ^ String.concat ", " (List.map string_of_vdecl vars) ^ "],\n" ^
  String.concat "\n" (List.map func_decl_s funcs) ^ ")"

(* "Pretty printed" version of the AST, meant to generate a MicroC program
   from the AST.  These functions are only for pretty-printing (the -a flag)
   the AST and can be removed. *)
let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | Id(s, e) -> s ^ string_of_expr e
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      (match o with
	Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
      | Exp -> "^" | Equal -> "==" | Neq -> "!="
      | Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">=") ^ " " ^
      string_of_expr e2
  | String_literal(s) -> "String Lit: " ^ s ^ "."
  | Float_literal(f) -> "Float lit: " ^ string_of_float f ^ "."
  | Bool_literal(b) -> "Bool lit: " ^ string_of_bool b ^ "."
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Basic_Op(b, s, i, a, e) -> 
			(match b with Signal -> "Signal" | Array -> "Array") ^ " " ^
			s ^ (match b with Signal -> "[time - " | Array -> "[") 
			^ string_of_int i ^ "] = " ^ string_of_expr e
  | Noexpr -> ""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s
  | Print(s) -> "Printing: " ^ string_of_expr s ^ "."
  | Basic_Dec(s, b, i) -> (match b with Signal -> "Signal" | Array -> "Array") ^ " " ^ s ^ " " ^ string_of_int i ^ "."
  | Prim_Assign(v) -> string_of_var v

let string_of_fdecl fdecl =
  (match fdecl.is_stream with true -> "Stream" | false -> "")^ " type: "
  ^ string_of_type fdecl.ret_type ^ " " 
  ^fdecl.fname ^ "(" ^ String.concat ", " (List.map string_of_formal fdecl.formals) ^ ")\n{\n" ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
